# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

from importlib import reload
import pytest

from MDAnalysis.analysis import hole2
from MDAnalysisTests.util import import_not_available


@pytest.mark.skipif(
    import_not_available("mdahole2"),
    reason="Test skipped because mdahole2 is not found",
)
def test_moved_to_mdakit_warning():
    wmsg = "MDAnalysis.analysis.hole2 is deprecated"
    with pytest.warns(DeprecationWarning, match=wmsg):
        reload(hole2)


@pytest.mark.skipif(
    not import_not_available("mdahole2"),
    reason="Test skipped because mdahole2 is found",
)
def test_install_mdakit_warning():
    wmsg = "Please install"
    with pytest.warns(UserWarning, match=wmsg):
        reload(hole2)
