# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda

from MDAnalysisTests.topology.base import ParserBase

from MDAnalysisTests.datafiles import (
    PDBQT_input,  # pdbqt_inputpdbqt.pdbqt
    PDBQT_tyrosol,  # tyrosol.pdbqt.bz2
)

from numpy.testing import assert_allclose


class TestPDBQT(ParserBase):
    parser = mda.topology.PDBQTParser.PDBQTParser
    ref_filename = PDBQT_input
    expected_attrs = [
        "ids",
        "names",
        "charges",
        "types",
        "altLocs",
        "resids",
        "resnames",
        "segids",
        "chainIDs",
        "record_types",
        "icodes",
        "occupancies",
        "tempfactors",
    ]

    guessed_attrs = ["masses"]
    expected_n_atoms = 1805
    expected_n_residues = 199  # resids go 2-102 then 2-99
    expected_n_segments = 2  # res2-102 are A, 2-99 are B

    def test_guessed_masses(self, filename):
        u = mda.Universe(filename)
        assert_allclose(
            u.atoms.masses[:7], [14.007, 0.0, 0.0, 12.011, 12.011, 0.0, 12.011]
        )


def test_footnote():
    """just test that the Universe is built even in the presence of a
    footnote before the charges.
    """
    mda.Universe(PDBQT_tyrosol)
